class LightWeightEdgeList:
    def __init__(self):
        self._edge_list: list = []
        self._topologies: list = []
        self._joint_degrees: list = []
        self._motif_id: list = []

    @property
    def edge_list(self) -> list:
        return self._edge_list

    @edge_list.setter
    def edge_list(self, value: list) -> None:
        self._edge_list = value

    @property
    def topologies(self) -> list:
        return self._topologies

    @topologies.setter
    def topologies(self, value: list) -> None:
        self._topologies = value

    @property
    def joint_degrees(self) -> list:
        return self._joint_degrees

    @joint_degrees.setter
    def joint_degrees(self, value: list) -> None:
        self._joint_degrees = value

    @property
    def motif_id(self) -> list:
        return self._motif_id

    @motif_id.setter
    def motif_id(self, value: list) -> None:
        self._motif_id = value



from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeEmpirical(JointDegree):

    _type: str = JointDegreeType.EMPIRICAL

    def __init__(self, params: dict):
        self._empirical_jds: list = []
        try:
            self._motif_sizes = params[JointDegreeNames.MOTIF_SIZES]
            self._empirical_jds = params[JointDegreeNames.JDS]
        except Exception as e:
            raise (f"Error instantiating {self.__class__.__name__}: {e}")
        self.create_jdd()

    def create_jdd(self) -> None:
        self.convert_jds_to_jdd(self._empirical_jds)

    @property
    def empirical_jds(self) -> list:
        return self._empirical_jds

    @empirical_jds.setter
    def empirical_jds(self, value: list) -> None:
        self._empirical_jds = value



import numpy as np


def poisson(kmean: float) -> callable:
    """
    Implements a poisson distribution
    :param kmean: mean of poisson distribution
    :returns p: Callable
    """

    def p(k: int) -> float:
        return np.exp(-kmean) * pow(kmean, k) / np.math.factorial(k)

    return p



from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_factory import JointDegreeFactory
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeDistribution:
    """
    Determines the type of JointDegree subclass to return. Expects `params`
    dict to contain key `joint_degree_type'.
    :method load_joint_degree: returns a subclass of ABC `JointDegree`. This
    will raise an error if the params dict does not contain the required keys.
    """

    @staticmethod
    def load_joint_degree(params: dict) -> JointDegree:
        try:
            input_type = JointDegreeType(params[JointDegreeNames.JOINT_DEGREE_TYPE])
        except Exception as e:
            raise (f"Error instantiating JointDegreeDistribution: {e}")
        loader: JointDegree = JointDegreeFactory.resolve_joint_degree(
            input_type, params
        )
        loader.create_jdd()
        return loader



from enum import Enum


class GCMAlgorithmNames(Enum):
    GCM_TYPE = "GCM_type"
    MOTIF_SIZES = "motif_sizes"
    BUILD_FUNCTIONS = "build_functions"
    EDGE_NAMES = "edge_names"
    MOTIF_INDICES = "motif_indices"
    MOTIF_ID = "motif_id"



import networkx as nx


class Network:
    def __init__(self):
        self._G = nx.Graph()

    def add_edge(self, e: tuple):
        """
        Adds an edge to the networkx
        :param e tuple: edge
        """
        self._G.add_edge(*e)

    def add_edges_from(self, edges: list) -> None:
        """
        Adds edges from list of tuples (int,int) to the edge list.
        :param edges: list of tuples of ints.
        """
        self._G.add_edges_from(edges)

    def find_cliques(self):
        """
        Returns all maximal cliques
        """
        return list(nx.find_cliques(self._G))

    def remove_edge(self, i: int, j: int) -> None:
        """
        Removes edge (i,j) from G. With nx, will also remove (j,i).
        """
        try:
            self._G.remove_edge(i, j)
        except nx.NetworkXError:
            return

    def has_edges(self) -> bool:
        """
        True if graph has edges remaining.
        """
        return len(self._G.edges()) > 0

    @property
    def G(self) -> nx.Graph:
        return self._G

    @G.setter
    def G(self, value: nx.Graph) -> None:
        self._G = value



from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeManual(JointDegree):

    _type: str = JointDegreeType.MANUAL

    def __init__(self, params: dict):
        try:
            self._jdd = params[JointDegreeNames.JDD]
            self._motif_sizes = params[JointDegreeNames.MOTIF_SIZES]
        except Exception as e:
            raise (f"Error instantiating {self.__class__.__name__}: {e}")
        self.create_jdd()

    def create_jdd(self) -> None:
        return



from gcmpy.network.network import Network
from gcmpy.gcm_algorithm.gcm_algorithm import GCMAlgorithm
from gcmpy.gcm_algorithm.gcm_algorithm_fast import GCMAlgorithmFast
from gcmpy.network.edge_list_to_network import EdgeListToNetwork
from gcmpy.names.gcm_algorithm_names import GCMAlgorithmNames


class GCMAlgorithmNetwork(GCMAlgorithm):
    def random_clustered_graph(self, jds: list) -> Network:
        params = {}
        params[GCMAlgorithmNames.MOTIF_SIZES] = self._motif_sizes
        params[GCMAlgorithmNames.BUILD_FUNCTIONS] = self._build_functions
        params[GCMAlgorithmNames.EDGE_NAMES] = self._edge_names
        CEdgeList = GCMAlgorithmFast(params).random_clustered_graph(jds)

        return EdgeListToNetwork.convert(CEdgeList)



from enum import Enum


class JointDegreeType(Enum):
    MANUAL = "manual"
    EMPIRICAL = "empirical"
    JOINT_FUNCTION = "function"
    MARGINAL = "marginal"
    SPLIT_DEGREE = "split_degree"
    DELTA = "delta"
    COVER = "cover"
    UNDEFINED = "undefined"



from itertools import combinations


def clique_motif(vertices: list) -> list:
    """
    Accepts list of ints and creates all possible pairs which it returns as a list of tuples of int pairs
    :param vertices: list of vertices (int)
    :returns: edge list as list of tuples (int, int)
    """
    return list(combinations(vertices, 2))



class JointDegreeFromExcess:
    @staticmethod
    def invert_single(qk: dict, i: int) -> dict:
        """
        Invert a excess joint degree distribution to a joint
        degree distribution.
        :param qk: dict excess joint degree distribution
        :param i: index of current topology in joint degree tuple
        :returns dict: joint degree distribution
        """
        P = {}
        bottom = sum(
            [(qk[joint_excess] / (joint_excess[i] + 1)) for joint_excess in qk]
        )
        for joint_excess in qk:
            top = qk[joint_excess] / (joint_excess[i] + 1)
            joint_degree = list(joint_excess)
            joint_degree[i] += 1
            P[tuple(joint_degree)] = top / bottom
        return P

    @staticmethod
    def observations_from_dict(qks: dict[dict], keys: list[str]) -> dict[dict]:
        P_observations = {}
        for i, key in enumerate(keys):
            P_observations[key] = JointDegreeFromExcess.invert_single(qks[key], i)
        return P_observations

    @staticmethod
    def get_joint_degree_distribution(qks: dict[dict], keys: list[str]) -> dict:
        p_obs: dict[dict] = JointDegreeFromExcess.observations_from_dict(qks, keys)
        # get set of common keys across all observations
        p_obs_list = [p_obs[topology] for topology in p_obs]
        common_keys = list(set.intersection(*map(set, p_obs_list)))

        if not common_keys:
            raise "Error: JointDDFromExcess - no common keys found across observations"

        # choose a common key and a reference topology to scale to
        common_key = common_keys[0]
        choesn_topology = "2-clique"

        # scale all observations to the chosen value
        base_value = p_obs[choesn_topology][common_key]
        for topology in p_obs:
            if choesn_topology == topology:
                continue
            scale_factor = base_value / p_obs[topology][common_key]
            for key in p_obs[topology]:
                p_obs[topology][key] *= scale_factor

        # merge keys
        P = {}
        for topology in p_obs:
            P.update(p_obs[topology])

        # renormalise
        total = sum(P.values())
        for k in P:
            P[k] /= total

        return P



from abc import ABC, abstractmethod
from typing import Any, Generator

from gcmpy.names.gcm_algorithm_names import GCMAlgorithmNames


class GCMAlgorithm(ABC):
    """
    Abstract base class for GCM algorithm. Expects a parameter dict with keys:
    :param motif_sizes: list of ints that indicate the number of vertices in each motif
    :param build_functions: callbacks that accept list of vertices and return edges
    :param edge_names: list of names for edge topologies
    """

    def __init__(self, params: dict):
        self._motif_sizes: list = []  # list of number of vertices in each motif
        self._build_functions: list = []  # list of callbacks for motif construction
        self._edge_names: list = []  # list of edge topology names

        try:
            self._motif_sizes = params[GCMAlgorithmNames.MOTIF_SIZES]
            self._build_functions = params[GCMAlgorithmNames.BUILD_FUNCTIONS]
            self._edge_names = params[GCMAlgorithmNames.EDGE_NAMES]
        except Exception as e:
            raise (f"Error in {self.__class__.__name__}: {e}")

    def __new__(cls, *args, **kwargs):
        if cls is GCMAlgorithm:
            raise TypeError(
                "The GCMAlgorithm class is abstract and may not be instantiated"
            )
        return object.__new__(cls)

    @abstractmethod
    def random_clustered_graph(self, jds: list) -> Any:
        raise NotImplementedError(
            "Error attempting to call virtual method on GCMAlgorithm: random_clustered_graph"
        )

    def infinite_sequence(self) -> Generator:
        num = 0
        while True:
            yield num
            num += 1



from itertools import product

from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeFunction(JointDegree):
    """
    Multivariate function to evaluate the probability of given joint degree from an analytical
    source. Note, callable self._fp signature must accept a joint degree tuple and return a float.
    :param fp: callback
    :param motif_sizes: list of ints for number of vertices in each motif
    :param hi_lo_degree_bounds: list of tuples (int,int) for kmin,kmax per topology
    :param n_samples: number of samples if not direct
    """

    _type: str = JointDegreeType.JOINT_FUNCTION

    def __init__(self, param: dict):
        self._fp: callable = None
        self._low_high_degree_bounds: tuple = (0, 50)
        try:
            self._motif_sizes = param[JointDegreeNames.MOTIF_SIZES]
            self._fp = param[JointDegreeNames.FP]
            self._low_high_degree_bounds = param[JointDegreeNames.LOW_HIGH_DEGREE_BOUND]
        except Exception as e:
            raise (f"Error instantiating {self.__class__.__name__}: {e}")

        self.create_jdd()

    def create_jdd(self) -> None:
        """
        Evaluates probability directly by generating all possible joint degrees.
        """
        # build list of lists of possible degrees in each dimension
        ks = [
            list(range(kmin, kmax + 1)) for kmin, kmax in self._low_high_degree_bounds
        ]
        # iterate all joint degrees and evaluate the joint degree
        for jd in list(product(*ks)):
            self._jdd[jd] = self._fp(jd)



from .gcm_algorithm import GCMAlgorithm
from .gcm_algorithm_factory import GCMAlgorithmFactory
from .gcm_algorithm_fast import GCMAlgorithmFast
from .gcm_algorithm_main import GCMAlgorithmMain
from .gcm_algorithm_custom_motifs import GCMAlgorithmCustomMotifs
from .gcm_algorithm_network import GCMAlgorithmNetwork
from .gcm_algorithm_types import GCMAlgorithmTypes



from gcmpy.motif_generators.cycle_motif import cycle_motif


def diamond_motif(vertices: list) -> list:
    """
    Accepts list of ints and creates edge pairs for a diamond motif of 4 vertices.
    :param vertices: list of vertices (int)
    :returns: edge list as list of tuples (int, int)
    """
    if len(vertices) < 4:
        raise ("Error during motif construction - diamond_motif")

    n0, n1, n2, n3 = vertices
    edges = cycle_motif(vertices)
    edges.append((n0, n2))
    edges.append((n1, n3))
    return edges



from itertools import combinations
from random import choice

from gcmpy.network.network import Network


def binom(n: int, r: int) -> int:
    """
    Binomial coefficient, nCr, aka the "choose" function
            n! / (r! * (n - r)!)
    :param n: n choose
    :param r: r
    :returns int: binomial coefficient.
    """
    p = 1
    for i in range(1, min(r, n - r) + 1):
        p *= n
        p //= i
        n -= 1
    return p


class EECC(Network):
    """
    Creates an EECC clique cover based on the algorithm developed in (1,2) for a graph
    in an edge_list format.

    ----- References -----

    1) Giulio Burgio, Alex Arenas, Sergio Gómez and Joan T. Matamalas:
       Network clique cover approximation to analyze complex contagions through group interactions,
       Comms. Phys. (2021) in press (arXiv:2101.03618)

    2) https://github.com/giubuig/DisjointCliqueCover.jl
    """

    def __init__(self):
        self._m0: int = 2
        super().__init__()

    def set_max_clique_size(self, m0: int) -> None:
        self._m0 = m0

    def limited_maximal_cliques(self) -> list:
        """
        Calculates the maximal cliques of a graph up to order m0. Any clique of order
        m>m0 is decomposed into its sub-cliques of order m0.
        :return list of maximal cliques up to size m0
        """
        indxs = []
        C = self.find_cliques()
        for c in range(len(C)):
            clique_size = len(C[c])
            if clique_size > self._m0:
                indxs.append(c)
                # sort by first element of sublist
                for nc in sorted(combinations(C[c], self._m0)):
                    C.append(nc)
            else:
                C[c] = sorted(C[c])

        idxs = set(range(len(C))) - set(indxs)

        C = [C[i] for i in idxs]
        C = [list(item) for item in set(tuple(row) for row in C)]

        # small difference in key for sort
        for i, c in enumerate(C):
            C[i] = sorted(c)

        return sorted(
            C, key=lambda x: (-len(x), x[0], x[1]) if len(x) > 1 else (-len(x), x[0], 0)
        )

    def compute_scores(
        self, C: list, EC: list, ord: list, r: list, indexes: list
    ) -> None:
        """
        Scores the cliques in list C in the network G and includes those of score zero in the
        cover EC. The higher the score
        :param C: set of maximal clique
        :param EC: EECC cover
        :param ord: List of cliques' order
        :param r: List of cliques' score
        :param indexes: List of indexes
        """
        num_cliques = len(C)
        for c in range(num_cliques):

            order = len(C[c])  # num vertives in clique
            C[c] = sorted(C[c])  # sort vertices
            ord[c] = order  # set size

            if order > 2:

                size = binom(order, 2)  # number of edges in clique of size order

                for i in range(order):  # for each vertex pair in clique (i.e. edges)
                    for j in range(i + 1, order):

                        f = 0
                        n = 0
                        # iterate all cliques and see if *this* edge (i,j)
                        # is a part of other cliques or not
                        while n <= num_cliques - 1:
                            if n != c and set([C[c][i], C[c][j]]).issubset(C[n]):
                                f = 1
                                break
                            else:
                                n += 1

                        if f == 1:
                            # increase the score for the overlapping edge
                            r[c] += 1.0 / size

            # if score is zero, add the clique to the cover
            if r[c] == 0:
                EC.append(C[c])
                indexes.append(c)

    def get_EECC(self) -> None:
        """
        Calculate the edge-disjoint edge clique cover (EECC) of a graph G considering
        cliques of order up to m0, according to the heuristic proposed in reference (1).
        :param G: graph
        :param m0: int for maximum order of the cliques to consider
        :return cover: A list containing the EECC
        """

        C = self.limited_maximal_cliques()
        num_cliques: int = len(C)
        ord: list = [0] * num_cliques
        r: list = [0.0] * num_cliques
        EC: list = []

        indexes_score0 = []
        self.compute_scores(C, EC, ord, r, indexes_score0)

        len_ = len(C)
        idxs = set(range(len_)) - set(indexes_score0)

        Ctemp = []
        ordtemp = []
        rtemp = []
        for i in idxs:
            Ctemp.append(C[i])
            ordtemp.append(ord[i])
            rtemp.append(r[i])

        C = Ctemp
        ord = ordtemp
        r = rtemp
        for c in range(len(EC)):
            order = len(EC[c])
            for i in range(order):
                for j in range(i + 1, order):
                    self.remove_edge(EC[c][i], EC[c][j])

        while self.has_edges():
            # find (one of) the largest clique(s) with the minimum score, include it in EECC

            # get smallest score
            min_r: float = min(r)

            # get indices of all smallest scoring cliques
            min_r_set = [idx for idx, element in enumerate(r) if element == min_r]

            # find size of largest clique of the smallest scoring cliques
            max_ord: int = -1
            for idx in min_r_set:
                if ord[idx] > max_ord:
                    max_ord = ord[idx]

            # construct a list of lowest-scoring, largest sized cliques
            indexes_to_sample = [idx for idx in min_r_set if ord[idx] == max_ord]

            # stochastic part of algorithm
            idx = choice(indexes_to_sample)

            # the selected clique to include in the EECC
            cli = C[idx]
            EC.append(cli)

            # remove the clique from the graph
            for i in range(max_ord):
                for j in range(i + 1, max_ord):
                    # assumes edges are ordered i < j
                    self.remove_edge(cli[i], cli[j])

            C = self.limited_maximal_cliques()

            # remove isolated vertices
            C = [c for c in C if len(c) > 1]
            ord = [0] * len(C)
            r = [0] * len(C)
            indexes_score0 = []

            self.compute_scores(C, EC, ord, r, indexes_score0)

            # Removing those cliques with score zero from C, r and G_
            len_ = len(C)
            idxs = set(range(len_)) - set(indexes_score0)

            Ctemp = []
            ordtemp = []
            rtemp = []
            for i in idxs:
                Ctemp.append(C[i])
                ordtemp.append(ord[i])
                rtemp.append(r[i])

            C = Ctemp
            ord = ordtemp
            r = rtemp
            for c in range(len(EC)):
                order = len(EC[c])

                for i in range(order):
                    for j in range(i + 1, order):
                        self.remove_edge(EC[c][i], EC[c][j])

        return sorted(EC, key=lambda x: (-len(x), x[0], x[1]))



import random
import networkx as nx
from logging import Logger

from gcmpy.names.tools_names import ToolsNames
from gcmpy.network.network import Network
from gcmpy.names.network_names import NetworkNames
from gcmpy.tools.joint_excess_joint_degree_matrices import (
    JointExcessJointDegreeMatrices,
)
from gcmpy.tools.proposal_edge import ProposalEdge
from gcmpy.tools.markov_chain_monte_carlo import MarkovChainMonteCarlo
from gcmpy.tools.joint_excess_joint_degree_keys_view import (
    JointExcessJointDegreeKeysView,
)
from gcmpy.tools.draw_set import DrawSet

class ErrorMarkovChainMonteCarloRewiring(Exception):
    ...


class MarkovChainMonteCarloRewiring(MarkovChainMonteCarlo):

    """
    Stochastically rewires `_network` so that the target correlations match `ejks`.
    """

    def __init__(self, params: dict):
        self._network: Network = None
        self._convergence_limit: int = None
        self._search_limit: int = 25
        self._ejks: JointExcessJointDegreeMatrices = None
        self._logger: Logger = Logger("MarkovChainMonteCarloRewiring")
        self._proposal_edges: list[ProposalEdge] = []
        self._proposal_count: int = 0
        self._proposals_accepted: int = 0
        self._acceptance_ratio: list = []
        try:
            self._network: Network = params[ToolsNames.NETWORK]
            self._ejks: JointExcessJointDegreeMatrices = params[ToolsNames.EJKS]
            if ToolsNames.CONVERGENCE_LIMIT in params:
                self._convergence_limit: int = params[ToolsNames.CONVERGENCE_LIMIT]
            else:
                # set a limit for the number of swaps before termination
                self._convergence_limit = 10 * self._network.G.edges()
            if ToolsNames.SEARCH_LIMIT in params:
                self._search_limit: int = params[ToolsNames.SEARCH_LIMIT]
        except Exception as e:
            raise ErrorMarkovChainMonteCarloRewiring(
                f"Error instantiating \
                {self.__class__.__name__}: {e}"
            )

    def get_all_edges(self, G: nx.Graph, u0: int, edge: tuple) -> list:
        """
        Returns all edges that have a given motif id as list of ordered tuples
        such that e = (u0,u1).
        :param G: nx.Graph
        :param e: edge in motif
        :returns list: all edges in motif involving focal vertex u0
        """
        es = []
        motif_id: int = G.edges[edge][NetworkNames.MOTIF_IDS]
        for e in G.edges(u0):
            if G.edges[e][NetworkNames.MOTIF_IDS] == motif_id:
                es.append((u0, self.get_other_vertex(u0, e)))
        return es

    def get_hashmap(self, G: nx.Graph, es: list) -> dict[list]:
        """
        Creates a hashmap of edges keyed by their topology.
        :param G: nx graph
        :param es: list of edges
        """
        hashmap_es: dict = {}
        for e in es:
            topology: str = G.edges[e][NetworkNames.TOPOLOGY]
            lst: list = hashmap_es.get(topology, [])
            lst.append(e)
            hashmap_es[topology] = lst
        return hashmap_es

    def is_edge_choice_suitable(
        self, G: nx.Graph, u0: int, v0: int, e0s: list, e1s: list
    ) -> bool:
        """
        Checks to see if the randomly chosen edge is suitable to be swapped. We
        assume that vertices e0[0] -> e1[0] and hence new edges will be created
        between u0 and all e1[1] neighbours; and, between v0 and e0[1] neighbours.
        Edges are not suitable if they don't have the same topology, do have the same
        motif id or if the target edge is already in the graph (possibly in another motif).

        :param G: nx.Graph
        :param u0: focal vertex on left
        :param v0: focal vertex on right
        :param e0s: edges involving focal vertex on left, ebunch
        :param e1s: edges involving focal vertex on right, ebunch
        :return bool: True if edges are OK to be swapped
        """
        if len(e0s) != len(e1s):
            # probably a mismatch during GCM algorithm, selfloops, multi-edges ...
            self._logger.debug(
                f"MarkovChainMonteCarlo - paired corners not equal size {len(e0s)} and {len(e1s)},\
                    {e0s} and {e1s}."
            )
            return False

        hashmap_e0s: dict = self.get_hashmap(G, e0s)
        hashmap_e1s: dict = self.get_hashmap(G, e1s)

        if hashmap_e0s.keys() != hashmap_e1s.keys():
            # check paired corners have same edge-topologies
            self._logger.debug(
                f"MarkovChainMonteCarlo - paired corners not equal topologies {hashmap_e0s} \
                    and {hashmap_e1s}"
            )
            return False

        for key in hashmap_e0s:
            if len(hashmap_e0s[key]) != len(hashmap_e1s[key]):
                # check paired corners have same number of each edge topology, residue
                # of GCM construction
                self._logger.debug(
                    f"MarkovChainMonteCarlo - paired corners not equal number of topologies {hashmap_e0s} \
                        and {hashmap_e1s}"
                )
                return False

        for e0, e1 in zip(e0s, e1s):
            if (
                G.edges[e0][NetworkNames.MOTIF_IDS]
                == G.edges[e1][NetworkNames.MOTIF_IDS]
            ):
                # other chosen vertex is in the same motif
                self._logger.debug(
                    "MarkovChainMonteCarlo - paired corners belong to same motif"
                )
                return False

        # check none of the potential target edges are already present
        # this checks all possible edges that *could* be made.
        for e0 in e0s:
            u1: int = self.get_other_vertex(u0, e0)
            topology: str = G.edges[e0][NetworkNames.TOPOLOGY]
            lst: list = hashmap_e1s[topology]
            for e1 in lst:
                v1: int = self.get_other_vertex(v0, e1)
                if G.has_edge(u0, v1) or G.has_edge(v0, u1):
                    self._logger.debug(
                        "MarkovChainMonteCarlo - target edges already in network"
                    )
                    return False
        return True

    def get_other_vertex(self, u: int, e: tuple) -> int:
        if e[0] == u:
            return e[1]
        elif e[1] == u:
            return e[0]
        else:
            raise ErrorMarkovChainMonteCarloRewiring(
                f"Error: MarkovChainMonteCarlo - get_other_node() vertex {u} not \
                in edge {e}"
            )

    def get_joint_excess_degree_key(self, G: nx.Graph, e: tuple, index: int) -> tuple:
        """
        Returns the joint excess degree key tuple for edge e in network G.
        :param G: nx.Graph
        :param e: edge
        :param index: index in joint degree of edge topology
        :returns key: joint excess joint degree key tuple
        """
        us_jd: list = [list(G.nodes[u][NetworkNames.JOINT_DEGREE]) for u in e]

        us_excess_jd: list = []
        for jd in us_jd:
            jd[index] -= 1
            us_excess_jd.append(tuple(jd))

        return us_excess_jd[0] + us_excess_jd[1]

    def get_swapped_joint_excess_degree_key(
        self, G: nx.Graph, e0: tuple, e1: tuple, u0: int, v0: int, index: int
    ) -> JointExcessJointDegreeKeysView:
        """
        Returns the swapped joint excess degree key tuple for two edges e0, e1 in network G.
        :param G: nx.Graph
        :param e0: edge
        :param e0: edge
        :param u0: left vertex
        :param v0: right vertex
        :param index: index in joint degree of edge topology
        :returns JointExcessJointDegreeKeysView: object to view excess joint degree tuples
        """
        u1: int = self.get_other_vertex(u0, e0)
        v1: int = self.get_other_vertex(v0, e1)

        # pull the joint degrees of the 4 vertices
        us: list = [u0, u1, v0, v1]
        us_jd: list = [list(G.nodes[u][NetworkNames.JOINT_DEGREE]) for u in us]

        # grab their excess degrees in topology `index`
        us_excess_jd: list = []
        for jd in us_jd:
            jd[index] -= 1
            us_excess_jd.append(tuple(jd))

        return JointExcessJointDegreeKeysView(us_excess_jd)

    def append_proposal_edges(
        self, G: nx.Graph, u0: int, old_edge: tuple, new_edge: tuple
    ) -> None:
        """
        Appends edge `e` (tuple) to propsal edges for this trial. Retains the
        focal vertex as first element (u0, v1).
        :param G nx.Graph:
        :param old_edge tuple: edge
        :param new_edge tuple: edge
        """
        p = ProposalEdge()
        p._topology = G.edges[old_edge][NetworkNames.TOPOLOGY]
        p._motif_id = G.edges[old_edge][NetworkNames.MOTIF_IDS]
        v1: int = self.get_other_vertex(u0, new_edge)
        p._new_edge = (u0, v1)
        self._proposal_edges.append(p)

    @MarkovChainMonteCarlo.proposal_efficiency
    def swap_condition(
        self, G: nx.Graph, e0s: list, e1s: list, u0: int, v0: int
    ) -> bool:
        """
        Evaluates the Metropolis condition for the algorithm.

        .. math::
            \\pi=prod_{\\nu}\\prod_i \\frac{e_{tau,k_{u_1},\\nu,k_{v_i}}e_{\tau,k_{v_1},
            nu,k_{v_i}}}{e_{\tau,k_{u_1},\nu,k_{u_i}}e_{\tau,k_{v_1},\nu,k_{v_i}}}

        Ensures that corresponding edges are swapped by creating a topology hashmap
        of (v0,v1) edges so that subgraphs with different orbits swap correct edges.
        Once edges are paired during this step, they are saved in
        `self._proposal_edges`.

        :param G nx.Graph: network
        :param e0s list: list of left vertex edge's in motif, ebunch
        :param e1s list: list of right vertex' edge's in motif, ebunch
        :param u0: left vertex
        :param v0: right vertex
        """
        # create hashmap of (v0,v1) edges keyed by their topologies
        top: float = 1.0
        hashmap_e1s: dict = self.get_hashmap(G, e1s)

        # refresh the list of new edges for this trial
        self._proposal_edges = []

        # iterate (u0,u1) edges
        for e0 in e0s:
            topology: str = G.edges[e0][NetworkNames.TOPOLOGY]
            index: int = self._ejks.get_topology_index(topology)
            # pull partner with same topology from hashmap of e1s
            lst: list = hashmap_e1s[topology]

            try:
                # this is where the other edge is selected
                e1 = lst.pop()
            except IndexError as e:
                raise ErrorMarkovChainMonteCarloRewiring(
                    f"IndexError during swap condition {e}"
                )

            # cache out the new edges that this attempt is proposing to add, (u0, v1)
            # and (v0, u1), to self._proposal_edges
            self.append_proposal_edges(G, u0, e0, (u0, self.get_other_vertex(v0, e1)))
            self.append_proposal_edges(G, v0, e1, (v0, self.get_other_vertex(u0, e0)))

            # create an excess degree key view object
            key_view: JointExcessJointDegreeKeysView = (
                self.get_swapped_joint_excess_degree_key(G, e0, e1, u0, v0, index)
            )

            u0v1_key: tuple = key_view.get_u0v1()
            v0u1_key: tuple = key_view.get_v0u1()
            u0u1: tuple = key_view.get_u0u1()
            u1u0: tuple = key_view.get_u1u0()
            v0v1: tuple = key_view.get_v0v1()
            v1v0: tuple = key_view.get_v1v0()

            # test if swapped keys change anything, if not, silent return False.
            starting_keys: set = {u0u1, u1u0, v0v1, v1v0}
            if set([u0v1_key, v0u1_key]).issubset(starting_keys):
                return False

            try:
                top *= (
                    self._ejks.ejks[topology][u0v1_key]
                    * self._ejks.ejks[topology][v0u1_key]
                )
            except KeyError as e:
                # guard due to GCM algorithm honouring the handshaking lemma by inserting
                # additional motif. Ensure this is not statistically significant, unlikely
                # to fail here due to the previous checks ...
                self._logger.debug(
                    f"MarkovChainMonteCarlo - KeyError during swap condition numerator: {e}"
                )
                return False

            if top == 0.0:
                return False

        bottom: float = 1.0
        # unlike the numerator, the order of iteration of edges in e0s, e1s is unimportant
        # since we are multiplying the all together anyway. However, we should pull the
        # topology for each edge seperately.
        for e0, e1 in zip(e0s, e1s):

            left_topology: str = G.edges[e0][NetworkNames.TOPOLOGY]
            index: int = self._ejks.get_topology_index(left_topology)
            key_e0 = self.get_joint_excess_degree_key(G, e0, index)

            right_topology: str = G.edges[e1][NetworkNames.TOPOLOGY]
            index: int = self._ejks.get_topology_index(right_topology)
            key_e1 = self.get_joint_excess_degree_key(G, e1, index)

            try:
                bottom *= (
                    self._ejks.ejks[left_topology][key_e0]
                    * self._ejks.ejks[right_topology][key_e1]
                )
            except KeyError as e:
                # guard due to GCM algorithm honouring the handshaking lemma by inserting
                # additional motif. Ensure this is not statistically significant.
                self._logger.debug(
                    f"MarkovChainMonteCarlo - KeyError during swap condition denominator: {e}"
                )
                return False

        if bottom == 0.0:
            raise ErrorMarkovChainMonteCarloRewiring("swap_condition() divide by zero")

        value: float = (top + 0.0) / bottom
        if value > random.random():
            return True

        return False

    def rewire(self) -> nx.Graph:
        """
        Entrypoint to MCMC rewiring algorithm
        """
        # pull a copy of the network, careful still has edge data?
        G: nx.Graph = self._network.G.copy()

        number_of_edges: int = G.number_of_edges()

        # create a set based on AVL tree. Ensure to order
        # the edge to avoid duplicate entries for a given edge
        EdgeSet = DrawSet()
        for e in G.edges():
            EdgeSet.add(tuple(sorted(e)))

        convergence_count: int = 0
        while convergence_count <= self._convergence_limit:

            if convergence_count % 50 == 0 and self._proposal_count != 0:
                self._acceptance_ratio.append(
                    float(self._proposals_accepted) / float(self._proposal_count)
                )

            # pick an edge at random.
            e0: tuple = EdgeSet.draw()
            u0: int = e0[0]
            u_edges_in_motif: list = self.get_all_edges(G, u0, e0)

            search_count: int = 0
            while search_count <= self._search_limit:

                # choose another edge at random.
                e1: tuple = EdgeSet.draw()

                if (
                    G.edges[e1][NetworkNames.TOPOLOGY]
                    != G.edges[e0][NetworkNames.TOPOLOGY]
                ):
                    continue

                v0: int = e1[0]
                v_edges_in_motif: list = self.get_all_edges(G, v0, e1)
                if self.is_edge_choice_suitable(
                    G, u0, v0, u_edges_in_motif, v_edges_in_motif
                ):
                    break

                search_count += 1

            if search_count >= self._search_limit:
                self._logger.debug(
                    f"Error: MarkovChainMonteCarlo - rewire() could not find another edge to swap \
                    with {G.edges[e0]}"
                )
                continue

            # we now have two candidate corners of two seperate motifs of the same topology that
            # can be swapped next, evaluate the swap condition and then swap the edges.
            if self.swap_condition(G, u_edges_in_motif, v_edges_in_motif, u0, v0):

                convergence_count += 1

                # add the new proposal edges for both sides
                for pe in self._proposal_edges:
                    e: tuple = pe._new_edge
                    if G.has_edge(*e):
                        raise ErrorMarkovChainMonteCarloRewiring(
                            f"Edge {e} already present in network!"
                        )
                    G.add_edge(*e)
                    G.edges[e][NetworkNames.TOPOLOGY] = pe._topology
                    G.edges[e][NetworkNames.MOTIF_IDS] = pe._motif_id
                    EdgeSet.add(tuple(sorted(e)))

                # remove old edges
                for e0, e1 in zip(u_edges_in_motif, v_edges_in_motif):
                    G.remove_edge(*e0)
                    G.remove_edge(*e1)

                    # EdgeSet expects ordered tuples
                    EdgeSet.remove(tuple(sorted(e0)))
                    EdgeSet.remove(tuple(sorted(e1)))

                if G.number_of_edges() != number_of_edges:
                    raise ErrorMarkovChainMonteCarloRewiring(
                        f"Error: MarkovChainMonteCarlo - rewire() edge count not preserved\
                        by swap {number_of_edges} vs {G.number_of_edges()}"
                    )

        return G

    @property
    def network(self) -> Network:
        return self._network

    @network.setter
    def network(self, value: Network) -> None:
        self._network = value

    @property
    def convergence_limit(self) -> int:
        return self._convergence_limit

    @convergence_limit.setter
    def convergence_limit(self, value: int) -> None:
        self._convergence_limit = value

    @property
    def search_limit(self) -> int:
        return self._search_limit

    @search_limit.setter
    def search_limit(self, value: int) -> None:
        self._search_limit = value

    @property
    def ejks(self) -> JointExcessJointDegreeMatrices:
        return self._ejks

    @ejks.setter
    def ejks(self, value: JointExcessJointDegreeMatrices) -> None:
        self._ejks = value



from itertools import product
import numpy as np
import random

from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeMarginal(JointDegree):

    """
    Merge uncorrelated marginals in each topology from analytical data together to create
    self._jdd. If using a direct method, all possible joint degree tuples are evaluated;
    however, for large varience in the allowed degrees this method is slow. Instead, we can
    choose to sample the analytical functions by setting use_sampling which draws n_samples
    weighted samples from each marginal function and pieces them together.
    :param arr_fp: array of callbacks
    :param motif_sizes: list of ints for number of vertices in each motif
    :param hi_lo_degree_bounds: list of tuples (int,int) for kmin,kmax per topology
    :param use_sampling: bool to use sampling or direct approach
    :param n_samples: number of samples if not direct
    """

    _type: str = JointDegreeType.MARGINAL

    def __init__(self, params: dict):
        self._arr_fp: list[callable] = []
        self._low_high_degree_bounds: tuple = None
        self._use_sampling: bool = False
        self._n_samples: int = 100000

        try:
            self._motif_sizes = params[JointDegreeNames.MOTIF_SIZES]
            self._arr_fp = params[JointDegreeNames.ARR_FP]
            self._low_high_degree_bounds = params[
                JointDegreeNames.LOW_HIGH_DEGREE_BOUND
            ]
        except Exception as e:
            raise (f"Error instantiating {self.__class__.__name__}: {e}")

        # default not to use sampling method
        if JointDegreeNames.USE_SAMPLING in params:
            self._use_sampling = params[JointDegreeNames.USE_SAMPLING]
        if JointDegreeNames.N_SAMPLES in params:
            self._n_samples = params[JointDegreeNames.N_SAMPLES]
        self.create_jdd()

    def create_jdd(self) -> None:
        if not self._use_sampling:
            self.create_jdd_directly()
        else:
            self.create_jdd_by_sampling()

    def create_jdd_directly(self) -> None:
        """
        Create self._jdd by enumerating probability of all possible joint degrees from
        analytical functions.
        """
        self._jdd = dict((key, 0.0) for key in self.generate_all_joint_degrees())
        for key in self._jdd:
            self._jdd[key] = self.evaluate_prob_of_joint_degree(key)
        self.normalise_jdd()

    def generate_all_joint_degrees(self) -> list:
        """
        Generate all possible joint degrees from a range of min/max degree of each motif.
        :return jd: list of joint degrees
        """
        ks = []
        for kmin, kmax in self._low_high_degree_bounds:
            ks.append([k for k in range(kmin, kmax)])
        return list(product(*ks))

    def evaluate_prob_of_joint_degree(self, joint_degree: list) -> float:
        """
        Evaluate the joint probability of a joint degree using function callbacks.
        :param joint_degree: list of joint degreees
        :returns prod: probability of joint degree
        """
        prod: float = 1.0
        for i, deg in enumerate(joint_degree):
            prod *= self._arr_fp[i](deg)
        return prod

    def draw_from_analytical_joint(self) -> list:
        """Draw `self._n_samples' from a marginal distribution `self._arr_fp' along each dimension.
        :returns jds: samples from marginal distributions"""
        ret = []
        for i in range(len(self._low_high_degree_bounds)):
            kmin, kmax = self._low_high_degree_bounds[i]
            ks = [k for k in range(kmin, kmax + 1)]  # possible degrees
            pks = [self._arr_fp[i](k) for k in ks]  # degree weights
            ret.append(
                random.choices(ks, pks, k=self._n_samples)
            )  # sample this dimension
        return [
            tuple(jd) for jd in np.column_stack(ret).tolist()
        ]  # return sampled degrees

    def create_jdd_by_sampling(self) -> None:
        """
        Draws jds samples from analytical marginal functions and converts to
        a joint degree distribution.
        """
        self.convert_jds_to_jdd(self.draw_from_analytical_joint())



from enum import Enum


class ToolsNames(Enum):
    NETWORK = "network"
    JDD = "jdd"
    EDGE_NAMES = "edge_names"
    EJKS = "ejks"
    EXCESS_DEGREE_KEYS = "excess_degree_keys"
    CONVERGENCE_LIMIT = "convergence_limit"
    SEARCH_LIMIT = "search_limit"



import random
from itertools import chain, repeat, starmap

from gcmpy.gcm_algorithm.gcm_algorithm import GCMAlgorithm
from gcmpy.network.edge_list import LightWeightEdgeList
from gcmpy.names.gcm_algorithm_names import GCMAlgorithmNames


class GCMAlgorithmCustomMotifs(GCMAlgorithm):
    def __init__(self, params: dict):
        self._motif_indices: list = []
        try:
            self._motif_indices = params[GCMAlgorithmNames.MOTIF_INDICES]
            super().__init__(params)
        except Exception as e:
            raise (f"Error in {self.__class__.__name__}: {e}")

    def partition(self, lst: list, n: int) -> list:
        """
        Yield successive n-sized partitions from list.
        Silent if lst % n != 0.
        """
        return [lst[i : i + n] for i in range(0, len(lst), n)]

    def random_clustered_graph(self, jds: list) -> LightWeightEdgeList:

        stubs = [
            list(chain.from_iterable(starmap(repeat, r)))
            for r in map(enumerate, zip(*jds))
        ]

        # shuffle each stub list
        for k_list in stubs:
            random.shuffle(k_list)

        # create list for edges and add joint degree sequence
        EdgeList = LightWeightEdgeList()
        EdgeList.joint_degrees = jds

        # split the stub lists into partitions of equal
        # size to the number of that topology required to
        # construct the motif
        partitions: list[list] = []
        for i, k_list in enumerate(stubs):
            partitions.append(self.partition(k_list, self._motif_sizes[i]))

        # self._motif_vertices is a list of lists that contain
        # the indices of the joint degree slots required to construct
        # the motif.

        gen = self.infinite_sequence()

        # for each motif type
        for j, motif_indexes in enumerate(self._motif_indices):

            # for each motif of that type
            kk: int = motif_indexes[0]
            num_motifs = (0.0 + len(stubs[kk])) / self._motif_sizes[kk]
            for k in range(int(num_motifs)):

                vertices: list = []
                # for each orbit in the motif
                for index in motif_indexes:
                    vertices.append(partitions[index].pop())

                vertices = [item for sublist in vertices for item in sublist]

                # build the motif edges from the vertices
                es: list = self._build_functions[j](vertices)

                # get the motif id
                id = next(gen)
                EdgeList.motif_id.extend([id] * len(es))

                if len(es) == 2:
                    # if 2-clique tuple annoyingly unpacks ... so re-pack it
                    EdgeList.edge_list.extend([es])
                    EdgeList.topologies.extend([self._edge_names[j]()])

                else:
                    EdgeList.edge_list.extend(es)
                    EdgeList.topologies.extend(self._edge_names[j]())

        return EdgeList



from .joint_degree import JointDegree
from .joint_degree_type import JointDegreeType
from .joint_degree_factory import JointDegreeFactory
from .joint_degree_distribution import JointDegreeDistribution



from gcmpy.names.joint_degree_names import JointDegreeNames
from gcmpy.names.network_names import NetworkNames
from gcmpy.names.tools_names import ToolsNames
from gcmpy.names.gcm_algorithm_names import GCMAlgorithmNames

from gcmpy.network.network import Network
from gcmpy.network.edge_list import LightWeightEdgeList
from gcmpy.network.edge_list_to_network import EdgeListToNetwork
from gcmpy.network.network_to_edge_list import NetworkToEdgeList

from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.joint_degree.joint_degree_factory import JointDegreeFactory
from gcmpy.joint_degree.joint_degree_distribution import JointDegreeDistribution
from gcmpy.joint_degree.joint_degree_loaders.joint_degree_manual import JointDegreeManual
from gcmpy.joint_degree.joint_degree_loaders.joint_degree_marginal import JointDegreeMarginal
from gcmpy.joint_degree.joint_degree_loaders.joint_degree_empirical import JointDegreeEmpirical
from gcmpy.joint_degree.joint_degree_loaders.joint_degree_cover import JointDegreeCover
from gcmpy.joint_degree.joint_degree_loaders.joint_degree_delta import JointDegreeDelta
from gcmpy.joint_degree.joint_degree_loaders.joint_degree_split_degree import JointDegreeSplitDegree
from gcmpy.joint_degree.joint_degree_loaders.joint_degree_function import JointDegreeFunction

from gcmpy.motif_generators.clique_motif import clique_motif
from gcmpy.motif_generators.cycle_motif import cycle_motif
from gcmpy.motif_generators.diamond_motif import diamond_motif

from gcmpy.gcm_algorithm.gcm_algorithm import GCMAlgorithm
from gcmpy.gcm_algorithm.gcm_algorithm_factory import GCMAlgorithmFactory
from gcmpy.gcm_algorithm.gcm_algorithm_fast import GCMAlgorithmFast
from gcmpy.gcm_algorithm.gcm_algorithm_main import GCMAlgorithmMain
from gcmpy.gcm_algorithm.gcm_algorithm_custom_motifs import GCMAlgorithmCustomMotifs
from gcmpy.gcm_algorithm.gcm_algorithm_network import GCMAlgorithmNetwork
from gcmpy.gcm_algorithm.gcm_algorithm_types import GCMAlgorithmTypes

from gcmpy.distributions.exponential import exponential
from gcmpy.distributions.poisson import poisson
from gcmpy.distributions.power_law import power_law
from gcmpy.distributions.scale_free_cut_off import scale_free_cut_off

from gcmpy.covers.eecc import EECC
from gcmpy.covers.mpcc import MPCC

from gcmpy.tools.average_joint_degree_from_jdd import AverageJointDegreeFromJDD
from gcmpy.tools.joint_degree_distribution_from_network import JointDegreeDistributionFromNetwork
from gcmpy.tools.joint_degree_from_excess import JointDegreeFromExcess
from gcmpy.tools.joint_excess_from_ejk import JointExcessFromEjk
from gcmpy.tools.joint_excess_from_jdd import JointExcessfromJDD
from gcmpy.tools.joint_excess_joint_degree import JointExcessJointDegree
from gcmpy.tools.joint_excess_joint_degree_matrices import JointExcessJointDegreeMatrices
from gcmpy.tools.markov_chain_monte_carlo_rewiring import MarkovChainMonteCarloRewiring
from gcmpy.tools.draw_set import DrawSet
from gcmpy.tools.bond_percolate import bond_percolate

from gcmpy.message_passing.message_passing_mixin import MessagePassingMixin
from gcmpy.message_passing.message_passing import MessagePassing
from gcmpy.message_passing.number_connected_graphs import number_of_connected_graphs, QQ, Q



from enum import Enum


class JointDegreeNames(Enum):
    JOINT_DEGREE_TYPE = "joint_degree_type"
    COVER = "cover"
    TARGET_K = "target_k"
    FP = "fp"
    PROBS = "probs"
    MOTIF_SIZES = "motif_sizes"
    LOW_HIGH_DEGREE_BOUND = "low_high_degree_bound"
    JDS = "jds"
    JDD = "jdd"
    ARR_FP = "arr_fp"
    USE_SAMPLING = "use_sampling"
    N_SAMPLES = "n_samples"



import random
import networkx as nx

def bond_percolate(g: nx.Graph, phi: float) -> float:
    """
    Performs bond percolation with occupancy probability `phi`
    :param g: nx graph to percolate (original)
    :param phi: bond occupancy probability
    :returns float S: size of the GCC
    """
    G: nx.Graph = g.copy()
    es: list = [e for e in G.edges() if random.random() > phi]
    G.remove_edges_from(es)
    Gcc: list = sorted(nx.connected_components(G), key=len, reverse=True)
    return float(len(Gcc[0]))/G.order()


from gcmpy.network.edge_list import LightWeightEdgeList
from gcmpy.network.network import Network
from gcmpy.names.network_names import NetworkNames


class NetworkToEdgeList:
    @staticmethod
    def convert(network: Network) -> LightWeightEdgeList:
        model = LightWeightEdgeList()
        model.edge_list = list(network.G.edges())
        model.joint_degrees = [
            network.G.nodes[n][NetworkNames.JOINT_DEGREE]
            for n in range(len(network.G.nodes()))
        ]
        model.topologies = [
            network.G.edges[e][NetworkNames.TOPOLOGY] for e in network.G.edges()
        ]
        model.motif_id = [
            network.G.edges[e][NetworkNames.MOTIF_IDS] for e in network.G.edges()
        ]
        return model



from gcmpy.tools.joint_excess_joint_degree_matrices import (
    JointExcessJointDegreeMatrices,
)


class JointExcessFromEjk:
    """
    Extracts the joint excess degree distributions from a
    list of joint excess joint degree distributions. This is
    done by summing the rows of the ejk matrices.
    """

    @staticmethod
    def get_excess_joint_distributions(
        ejks: JointExcessJointDegreeMatrices,
    ) -> dict[dict]:

        if len(ejks._ejks) != len(ejks.excess_degree_keys):
            raise ("Error in JointExcessDDFromEjk: incorrect ejk or keys provided")

        qks: dict = {}
        for key in ejks._ejks:
            ejk = ejks._ejks[key]
            keys = ejks._excess_degree_keys[key]
            q = {}
            for left_key in keys:
                for right_key in keys:
                    q[left_key] = q.get(left_key, 0.0) + ejk[left_key + right_key]
            qks[key] = q
        return qks



from gcmpy.names.tools_names import ToolsNames


class JointExcessJointDegreeMatrices:
    def __init__(self, params=None):
        self._ejks: dict[dict] = {}  # dict of ejk dicts/matrices keyed by edge topology
        self._excess_degree_keys: dict[
            list
        ] = {}  # dict of lst of excess deg-tuples keyed by edge topology
        self._topology_names: list[str] = []  # list of str of topology names, ordered

        if params is not None:
            self._ejks = params[ToolsNames.EJKS]
            self._topology_names = params[ToolsNames.EDGE_NAMES]
            self.get_excess_degree_keys()

    def get_topology_index(self, topology: str) -> int:
        for i, name in enumerate(self._topology_names):
            if name == topology:
                return i
        raise f"Error: JointExcessJointDegreeMatrices - no topology {name} in \
            names {self._topology_names}"

    def get_excess_degree_keys(self) -> None:
        """
        Method to get excess degree keys from ejk matrices for each topology.
        """
        self._excess_degree_keys: dict[list] = {}
        for topology in self._ejks:

            excess_key_set = set()
            for excess_key in self._ejks[topology]:

                A = list(excess_key)
                B = A[: len(A) // 2]
                C = A[len(A) // 2 :]
                excess_key_set.add(tuple(B))
                excess_key_set.add(tuple(C))

            self._excess_degree_keys[topology] = list(excess_key_set)

    @property
    def excess_degree_keys(self) -> dict:
        return self._excess_degree_keys

    @excess_degree_keys.setter
    def excess_degree_keys(self, value: dict) -> None:
        self._excess_degree_keys = value

    @property
    def ejks(self) -> dict:
        return self._ejks

    @ejks.setter
    def ejks(self, value: dict) -> None:
        self._ejks = value

    @property
    def topology_names(self) -> list:
        return self._topology_names

    @topology_names.setter
    def topology_names(self, value: list) -> None:
        self._topology_names = value



import networkx as nx

from gcmpy.network.edge_list import LightWeightEdgeList
from gcmpy.network.network import Network
from gcmpy.names.network_names import NetworkNames


class EdgeListToNetwork:
    @staticmethod
    def convert(edgelist: LightWeightEdgeList) -> Network:

        model = Network()
        model.G.add_edges_from(edgelist.edge_list)

        # create vertex attributes dict
        joint_degrees = {}
        for n, jd in enumerate(edgelist.joint_degrees):
            joint_degrees[n] = jd
        nx.set_node_attributes(model.G, joint_degrees, NetworkNames.JOINT_DEGREE)
        # create edge attributes dict
        topologies = {}
        motif_ids = {}
        for e, name, motif_id in zip(
            edgelist.edge_list, edgelist.topologies, edgelist.motif_id
        ):
            topologies[e] = name
            motif_ids[e] = motif_id

        nx.set_edge_attributes(model.G, topologies, NetworkNames.TOPOLOGY)
        nx.set_edge_attributes(model.G, motif_ids, NetworkNames.MOTIF_IDS)

        return model



import numpy as np


def scale_free_cut_off(alpha: float, kappa: float) -> callable:
    """
    Implements a scale free with exponential degree cutoff function. Limits
    to scale free for large degree cutoff. Undefined for k = 0.
    :param k: int degree
    :param alpha: float power law exponent
    :param kappa: float degree cutoff
    """

    def polylog(s: float, z: float) -> float:
        """
        Implements a polylogarithm function for real arguments, taking two floats
        :param s: base
        :param z: arg
        :returns float: polylogarithm
        """
        tol = +1e-06
        l = 0.0
        k = 1
        zk = z
        while 1:
            term = zk / k**s
            l += term
            if abs(term) < tol:
                break
            zk *= z
            k += 1
        return l

    C = polylog(alpha, np.exp(-1.0 / kappa))  # normalisation constant

    def p(k: int) -> float:
        return (pow((k + 0.0), -alpha) * np.exp(-(k + 0.0) / kappa)) / C

    return p



from gcmpy.gcm_algorithm.gcm_algorithm_types import GCMAlgorithmTypes
from gcmpy.gcm_algorithm.gcm_algorithm import GCMAlgorithm

from gcmpy.gcm_algorithm.gcm_algorithm_fast import GCMAlgorithmFast
from gcmpy.gcm_algorithm.gcm_algorithm_network import GCMAlgorithmNetwork
from gcmpy.gcm_algorithm.gcm_algorithm_custom_motifs import GCMAlgorithmCustomMotifs


class GCMAlgorithmFactory:
    @staticmethod
    def resolve_algorithm(type: GCMAlgorithmTypes, params: dict) -> GCMAlgorithm:
        if type == GCMAlgorithmTypes.FAST:
            return GCMAlgorithmFast(params)
        elif type == GCMAlgorithmTypes.NETWORK:
            return GCMAlgorithmNetwork(params)
        elif type == GCMAlgorithmTypes.MOTIFS:
            return GCMAlgorithmCustomMotifs(params)
        else:
            raise ("Error: unknown algorithm in GCMAlgorithmFactory: resolve_algorithm")



import networkx as nx
import itertools


def MPCC(G: nx.Graph):
    '''
    An edge-disjoint clique cover that preserves larger motifs within
    a network. This works by assigning labels to each edge of the form

    G.edges[e[0],e[1]]['clique'] = f'{len(c)}-{c}-{ID}'

    where c is a list of vertices in the clique and ID is a unique ID
    for the motif. For large networks the library call to
    `nx.enumerate_all_cliques` might be slow; however, we only need to
    call it once.

    Take care of self-loops in the network, as these will not be labelled.

    :param G nx.Graph: the network to cover.
    :returns: G a covered graph.
    '''
    g: nx.Graph = G.copy()
    cliques: list = sorted(list(nx.enumerate_all_cliques(g)), key=len, reverse=True)
    cover: list = []
    for c in cliques:
        skip: bool = False
        for e in itertools.combinations(c, 2):
            if not g.has_edge(e[0], e[1]):
                skip = True
                break
        if not skip:
            cover.append(c)
            g.remove_edges_from(list(itertools.combinations(c, 2)))

    clique_ID: int = itertools.count(0)
    for c in cover:
        ID: int = next(clique_ID)
        for e in itertools.combinations(c, 2):
            G.edges[e[0], e[1]]['clique'] = f'{len(c)}-{c}-{ID}'

    return G



from .joint_degree_manual import JointDegreeManual
from .joint_degree_marginal import JointDegreeMarginal
from .joint_degree_empirical import JointDegreeEmpirical
from .joint_degree_cover import JointDegreeCover
from .joint_degree_delta import JointDegreeDelta
from .joint_degree_split_degree import JointDegreeSplitDegree
from .joint_degree_function import JointDegreeFunction


import numpy as np


def exponential(a: float) -> callable:
    """
    Implemnts an exponential distribution.
    :param a: distribution parameter, a>0.
    :returns p: callable
    """

    def p(k: int) -> float:
        return (1 - np.exp(-a)) * np.exp(-a * k)

    return p



import random
from collections import Counter
from abc import ABC, abstractmethod


class JointDegree(ABC):

    """
    Abstract class whose purpose is to generate a joint degree
    distribution (jdd) dict through a variety of methods. A jdd
    can then be sampled at a later point during the GCM algorithm
    to create a joint degree sequence (jds).

    The abstract method `create_jdd' should be defined in a subclass
    and is the primary method that will be called to create an object
    to be sampled.
    """

    _type: str = ""

    def __init__(self):
        self._jdd: dict = None
        self._motif_sizes: list = None

    def __new__(cls, *args, **kwargs):
        if cls is JointDegree:
            raise TypeError(
                "The JointDegree class is abstract and may not be instantiated"
            )
        return object.__new__(cls)

    @abstractmethod
    def create_jdd(self) -> None:
        raise NotImplementedError(
            "Error attempting to call virtual method on JointDegree: create_joint_degree"
        )

    def handshaking_lemma(self, jds: list) -> list:
        """
        Samples a joint degree sequence from the joint degree distribution.
        Then ensures the handshaking lemma is satisfied by adding another
        subgraph to the network.
        """
        # ensure that the sum of the numbers is divisible by the motif size
        ntops = list(map(sum, zip(*jds)))
        for i, ntop in enumerate(ntops):
            if ntop % self._motif_sizes[i] != 0:
                # if not, round up to add one more motif to the network
                for j in range(self._motif_sizes[i] - ntop % self._motif_sizes[i]):
                    j = random.randrange(0, len(jds))
                    t = list(jds[j])
                    t[i] += 1
                    jds[j] = t
        return jds

    def sample_jds_from_jdd(self, N: int) -> list:
        """
        Chooses a list of `N' keys from `self._jdd' with replacement
        according to their weighting in the joint degree distribution.
        Checks the handshaking lemma is satisfied.
        :param N int: number of keys to choose
        :returns list: list of key choices
        """
        keys = list(self._jdd.keys())
        weights = list(self._jdd.values())
        jds = random.choices(population=keys, weights=weights, k=N)
        return self.handshaking_lemma(jds)

    def normalise_jdd(self) -> None:
        summation: float = sum(self._jdd.values())
        for key in self._jdd:
            self._jdd[key] /= summation

    def convert_jds_to_jdd(self, jds: list) -> None:
        n_samples: int = len(jds)
        self._jdd = {}
        d = Counter(jds)
        for k, v in d.items():
            self._jdd[k] = v / n_samples

    @property
    def jdd(self) -> dict:
        return self._jdd

    @jdd.setter
    def jdd(self, value: dict) -> None:
        self._jdd = value

    @property
    def motif_sizes(self) -> list:
        return self._motif_sizes

    @motif_sizes.setter
    def motif_sizes(self, value: list) -> None:
        self._motif_sizes = value



from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeCover(JointDegree):

    _type: str = JointDegreeType.COVER

    def __init__(self, params: dict):
        try:
            self._cover = params[JointDegreeNames.COVER]
        except Exception as e:
            raise (f"Error instantiating {self.__class__.__name__}: {e}")

        self._motif_sizes = sorted(list(set([len(c) for c in self._cover])))
        self.create_jdd()

    def create_jdd(self) -> None:

        vertex_ids = list(set([vertex for clique in self._cover for vertex in clique]))

        zero_index = 0
        if min(vertex_ids) != zero_index:
            zero_index = 1

        largest_clique = len(max(self._cover, key=len))

        jds = []
        for _ in range(len(vertex_ids)):
            jd = [0] * largest_clique
            jds.append(jd)

        for c in self._cover:
            clique_size = len(c)
            for vertex in c:
                jds[vertex - zero_index][clique_size - 1] += 1

        # iterate each column of the jds and record the index if all zeros
        indxs = [i for i, top in enumerate(zip(*jds)) if not any(top)]

        # use the indexes of the zero columns to remove
        for i in indxs:
            for jd in jds:
                del jd[i]

        # convert jds to jdd
        self.convert_jds_to_jdd(jds)

    @property
    def cover(self) -> list:
        return self._cover

    @cover.setter
    def cover(self, value: list) -> None:
        self._cover = value



from itertools import tee


def cycle_motif(vertices: list) -> list:
    """
    Accepts a list of ints and creates a cycle of adjacent vertex pairs. Creates two
    iterators and then increments one before zipping together to create tuples. Finally,
    connects the start and end of the chain toegher.
    :param vertices: list of vertices (int)
    :returns: edge list as list of tuples (int, int)
    """

    a, b = tee(vertices)
    _ = next(b, None)
    edges = list(zip(a, b))
    edges.append((vertices[0], vertices[-1]))
    return edges



class AverageJointDegreeFromJDD:
    @staticmethod
    def get_average_joint_degrees(jdd: dict) -> list:
        joint_degrees = list(jdd.keys())
        num_topologies = len(joint_degrees[0])
        average_degrees = [0.0] * num_topologies
        for joint_degree in joint_degrees:
            for index in range(num_topologies):
                average_degrees[index] += joint_degree[index] * jdd[joint_degree]
        return average_degrees



def power_law(alpha: float) -> callable:
    """
    Implements a power law distribution with exponent alpha. Undefined for k = 0.
    :param alpha: power law exponent
    :returns p: callable
    """

    def zeta(s: float) -> float:
        tol = +1e-06
        l = 0.0
        k = 1
        while 1:
            term = 1.0 / k**s
            l += term
            if abs(term) < tol:
                break
            k += 1
        return l

    C = zeta(alpha)

    def p(k: int) -> float:
        return pow(k, -alpha) / C

    return p



import random
from iteration_utilities import grouper
from itertools import chain, repeat, starmap

from gcmpy.gcm_algorithm.gcm_algorithm import GCMAlgorithm
from gcmpy.network.edge_list import LightWeightEdgeList


class GCMAlgorithmFast(GCMAlgorithm):
    def random_clustered_graph(self, jds: list) -> LightWeightEdgeList:

        stubs = [
            list(chain.from_iterable(starmap(repeat, r)))
            for r in map(enumerate, zip(*jds))
        ]

        # shuffle each stub list
        for k_list in stubs:
            random.shuffle(k_list)

        # create list for edges and add joint degree sequence
        EdgeList = LightWeightEdgeList()
        EdgeList.joint_degrees = jds

        gen = self.infinite_sequence()

        # for each topology list ...
        for k, k_list in enumerate(stubs):

            # iterate the degree list
            for vertices in grouper(k_list, self._motif_sizes[k]):
                # add the edges to the network using the builder callback
                es = self._build_functions[k](list(vertices))
                EdgeList.edge_list.extend(es)

                # add the edge names to a list
                EdgeList.topologies.extend([self._edge_names[k]] * len(es))

                # record the motif id
                id = next(gen)
                EdgeList.motif_id.extend([id] * len(es))

        # return the graph model as a LightWeightEdgeList
        return EdgeList



from functools import wraps


class MarkovChainMonteCarlo:

    """
    Base class for an MCMC experiment. Defines some decorators
    that yield the metadata of the Markov chain.
    """

    _proposal_count: int = 0
    _proposals_accepted: int = 0

    @classmethod
    def proposal_efficiency(self, func) -> callable:
        """
        Decorator to count the number of times a function is called
        and when it returns True. This is used to measure the efficiency
        of the Markov chain acceptance vs rejection.
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            self._proposal_count += 1
            result: bool = func(*args, **kwargs)
            if result:
                self._proposals_accepted += 1
            return result

        return wrapper



from .average_joint_degree_from_jdd import AverageJointDegreeFromJDD
from .joint_degree_distribution_from_network import JointDegreeDistributionFromNetwork
from .joint_degree_from_excess import JointDegreeFromExcess
from .joint_excess_from_ejk import JointExcessFromEjk
from .joint_excess_from_jdd import JointExcessfromJDD
from .joint_excess_joint_degree import JointExcessJointDegree
from .joint_excess_joint_degree_matrices import JointExcessJointDegreeMatrices
from .markov_chain_monte_carlo_rewiring import MarkovChainMonteCarloRewiring
from .draw_set import DrawSet


import random

class DrawSet(object):

    def __init__(self):
        self._edge_hashmap: dict = {}
        self._edges: list= []

    def __contains__(self, e: tuple) -> bool:
        return e in self._edge_hashmap

    def __iter__(self):
        return iter(self._edges)

    def __len__(self):
        return len(self._edges)

    def add(self, e: tuple) -> None:
        if e in self._edge_hashmap:
            return
        self._edges.append(e)
        self._edge_hashmap[e] = len(self._edges)-1

    def remove(self, e: tuple) -> None:
        position = self._edge_hashmap.pop(e)
        last_item = self._edges.pop()
        if position != len(self._edges):
            self._edges[position] = last_item
            self._edge_hashmap[last_item] = position

    def draw(self) -> tuple:
        return random.choice(self._edges)



from gcmpy.gcm_algorithm.gcm_algorithm_types import GCMAlgorithmTypes
from gcmpy.gcm_algorithm.gcm_algorithm_factory import GCMAlgorithmFactory
from gcmpy.gcm_algorithm.gcm_algorithm import GCMAlgorithm
from gcmpy.names.gcm_algorithm_names import GCMAlgorithmNames


class GCMAlgorithmMain:
    """
    Determines the type of GCM algorithm subclass to return. Expects `params`
    dict to contain key `GCM_type` among others additionally required
    by the subclasses.
    :method load_gcm_algorithm: returns the subclass of ABC `GCMAlgorithm`.
    This will raise an error if the params dict is not correctly set up.
    """

    @staticmethod
    def load_gcm_algorithm(params: dict) -> GCMAlgorithm:
        try:
            input_type = GCMAlgorithmTypes(params[GCMAlgorithmNames.GCM_TYPE])
            loader: GCMAlgorithm = GCMAlgorithmFactory.resolve_algorithm(
                input_type, params
            )
            return loader
        except Exception as e:
            raise (f"Error instantiating GeneralisedConfigurationModel: {e}")



import networkx as nx
from gcmpy.names.network_names import NetworkNames
from gcmpy.names.tools_names import ToolsNames
from gcmpy.tools.joint_excess_joint_degree_matrices import (
    JointExcessJointDegreeMatrices,
)


class JointExcessJointDegree:

    """
    Extracts the joint excess joint degree distribution from a network
    """

    def __init__(self, params: dict):
        self._topology_names: list[str] = None
        self._G: nx.Graph = None
        self._ejks = None
        self._num_edges: dict = {}  # number of edges of i-th topology
        self._degree_keys: list = []  # list of joint degree tuples
        self._excess_degree_keys = {}

        try:
            self._G: nx.Graph = params[ToolsNames.NETWORK]
            self._topology_names: list[str] = params[ToolsNames.EDGE_NAMES]
            self._degree_keys = set(
                [
                    tuple(self._G.nodes[n][NetworkNames.JOINT_DEGREE])
                    for n in self._G.nodes()
                ]
            )
        except Exception as e:
            raise (f"Error in {self.__class__.__name__}: {e}")

        self.resolve_excess_degree_keys()

    def resolve_excess_degree_keys(self) -> None:
        """
        Obtains the excess degree keys from the joint degree keys
        for each topology.
        """
        for i, topology in enumerate(self._topology_names):
            excess_keys_top_i = []
            for jd in self._degree_keys:
                if jd[i] > 0:
                    jd = list(jd)
                    jd[i] -= 1
                    excess_keys_top_i.append(tuple(jd))
            self._excess_degree_keys[topology] = list(set(excess_keys_top_i))

    def count_edge_types(self) -> None:
        """
        Enumerates the number of edges of the i-th topology.
        Assumes that the network edges have attribute `topology`.
        """
        # self._num_edges = {}
        for e in self._G.edges():
            topology: str = self._G.edges[e][NetworkNames.TOPOLOGY]
            self._num_edges[topology] = self._num_edges.get(topology, 0) + 1

    def get_ejk(self, i: int, name: str) -> dict:
        """
        Get mixing patterns for the i-th topology. Routine assumes that
        the network edges have attribute `topology` and that the vertices
        have attribute `joint_degree`.
        :param i int: index of topology i in joint degree
        :param name str: key for edge name
        :returns ejk dict: mixing patterns for this topology
        """
        ejk = {}
        for e in self._G.edges():
            if self._G.edges[e][NetworkNames.TOPOLOGY] == name:

                u, v = e

                u_joint_degree = list(self._G.nodes[u][NetworkNames.JOINT_DEGREE])
                v_joint_degree = list(self._G.nodes[v][NetworkNames.JOINT_DEGREE])

                u_joint_degree[i] -= 1
                u_joint_excess_degree = tuple(u_joint_degree)

                v_joint_degree[i] -= 1
                v_joint_excess_degree = tuple(v_joint_degree)

                key1 = u_joint_excess_degree + v_joint_excess_degree
                key2 = v_joint_excess_degree + u_joint_excess_degree

                if key1 == key2:
                    ejk[key1] = ejk.get(key1, 0) + (1.0 / (self._num_edges[name]))
                else:
                    ejk[key1] = ejk.get(key1, 0) + (1.0 / (self._num_edges[name]))
                    ejk[key2] = ejk.get(key2, 0) + (1.0 / (self._num_edges[name]))

        return ejk

    def get_ejks(self) -> JointExcessJointDegreeMatrices:
        """
        Get mixing patterns for all edge topologies
        :returns self._ejks: JointExcessJointDegreeMatrices.
        """
        self._ejks = JointExcessJointDegreeMatrices()
        self._ejks.excess_degree_keys = self._excess_degree_keys
        self._ejks.topology_names = self._topology_names

        self.count_edge_types()
        for i, topology in enumerate(self._topology_names):
            self._ejks.ejks[topology] = self.get_ejk(i, topology)
        return self._ejks



import networkx as nx
from gcmpy.names.network_names import NetworkNames


class JointDegreeDistributionFromNetwork:
    @staticmethod
    def get_joint_degree_distribution(G: nx.Graph) -> dict:
        """
        Records the joint degree from a given network. Expects
        the network vertices to have key 'joint_degree' which
        yields a tuple, this property is given when created using
        the GCMAlgorithm class.
        :param nx.Graph:
        :return dict: degree distribution
        """
        num_vertices = G.order()

        PK = {}
        for n in G.nodes():
            key = tuple(G.nodes[n][NetworkNames.JOINT_DEGREE])
            PK[key] = PK.get(key, 0) + (1.0 / num_vertices)

        return PK



class ProposalEdge:
    def __init__(self):
        self._topology: str = None
        self._motif_id: int = None
        self._new_edge: tuple = None

    @property
    def topology(self) -> str:
        return self._topology

    @topology.setter
    def topology(self, value: str) -> None:
        self._topology = value

    @property
    def motif_id(self) -> int:
        return self._motif_id

    @motif_id.setter
    def motif_id(self, value: int) -> None:
        self._motif_id = value

    @property
    def new_edge(self) -> tuple:
        return self._new_edge

    @new_edge.setter
    def new_edge(self, value: tuple) -> None:
        self._new_edge = value



from gcmpy.tools.average_joint_degree_from_jdd import AverageJointDegreeFromJDD


class JointExcessfromJDD:
    @staticmethod
    def get_joint_excess_distributions(jdd: dict) -> list[dict]:

        qks = []

        averages = AverageJointDegreeFromJDD.get_average_joint_degrees(jdd)

        joint_degrees = list(jdd.keys())
        num_topologies: int = len(joint_degrees[0])
        for index in range(num_topologies):
            q = {}
            for joint_degree in joint_degrees:
                _joint_degree = list(joint_degree)
                if _joint_degree[index] > 0:
                    _joint_degree[index] -= 1
                    q[tuple(_joint_degree)] = (
                        (_joint_degree[index] + 1) * jdd[joint_degree] + 0.0
                    ) / averages[index]
            qks.append(q)
        return qks

    @staticmethod
    def convert_list_qks_to_dict(qks_list: list[dict], keys: list[str]) -> dict[dict]:
        """
        Static method to convert a list of dicts to dict of dicts
        :param qks_list: list of excess degree distributions
        :param keys: list of topologies
        :returns dict: converted dict object
        """
        qks_dict = {}
        for key, qk in zip(keys, qks_list):
            qks_dict[key] = qk
        return qks_dict

    @staticmethod
    def convert_dict_qks_to_list(qks_dict: dict[dict], keys: list[str]) -> list[dict]:
        """
        Static method to convert a dict of dicts to list of dicts
        :param qks_dict: dict keyed by topology string
        :param keys list: list of topology keys (ordered)
        :returns list[dict]:
        """
        qks_list: list = []
        for key in keys:
            qks_list.append(qks_dict[key])
        return qks_list



# Setup for gcmpy
#
# Copyright (C) 2021 Peter Mann
#
# This file is part of gcmpy, generalised configuration model networks in Python.
#
# gcmpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# gcmpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gcmpy. If not, see <http://www.gnu.org/licenses/gpl.html>.


from setuptools import setup, find_packages

# the current version of the package
__version__ = "0.2.16"

with open('README.rst') as f:
    longDescription = f.read()

setup(
    name='gcmpy',
    version=__version__,
    description='Generalised Configuration Model random graphs in Python',
    long_description=longDescription,
    url='https://github.com/PeterStAndrews/gcmpy',
    author='Peter Mann',
    author_email='pm78@st-andrews.ac.uk',
    license='License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering'
    ],
    python_requires='>=3.8',
    packages=['gcmpy', 'gcmpy.covers','gcmpy.distributions','gcmpy.gcm_algorithm',
    'gcmpy.joint_degree','gcmpy.joint_degree.joint_degree_loaders', 'gcmpy.motif_generators','gcmpy.names','gcmpy.network', 'gcmpy.tools'],
    package_data={'gcmpy': ['py.typed']},
    zip_safe=False,
    install_requires=[
        "iteration_utilities==0.11.0",
        "networkx==2.6.3",
        "numpy==1.22.3"
    ]
)



from gcmpy.joint_degree.joint_degree import JointDegree
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeSplitDegree(JointDegree):

    _type: str = JointDegreeType.SPLIT_DEGREE

    def __init__(self, params: dict):
        self._fp: callable = None
        self._probs: list = []
        self._low_high_degree_bound: tuple = None

        try:
            self._fp = params[JointDegreeNames.FP]
            self._probs = params[JointDegreeNames.PROBS]
            self._motif_sizes = params[JointDegreeNames.MOTIF_SIZES]
            self._low_high_degree_bound = params[JointDegreeNames.LOW_HIGH_DEGREE_BOUND]
        except Exception as e:
            raise (f"Error instantiating {self.__class__.__name__}: {e}")

        self.create_jdd()

    def create_jdd(self) -> None:
        for k in range(self._low_high_degree_bound[0], self._low_high_degree_bound[1]):
            self.resolve_degree(k, self._fp(k))
        self.normalise_jdd()

    def get_valid_joint_degrees(self, remaining_degree: int, topology: int) -> list:
        """
        Returns a list of tuples by recursion. Only an ordered list of cliques are
        currently supported.
        :param remaining_degree: current free edges that can be partitioned
        :param topology: column index of joint degree tuple
        :returns list of joint degrees
        """
        if topology == 1:
            yield [remaining_degree]
        else:
            for i in range(0, remaining_degree // topology + 1):
                for row in self.get_valid_joint_degrees(
                    remaining_degree - i * topology, topology - 1
                ):
                    yield row + [i]

    def calc_prob_of_joint_degree(self, jd: tuple) -> float:
        """
        Calculates the probability of a joint degree from the input params.
        :param jd: joint degree
        :returns probability: float value
        """
        prod: float = 1.0
        for i, degree in enumerate(jd):
            prod *= pow(self._probs[i], (i + 1) * degree)
        return prod

    def resolve_degree(self, k: int, prob_overall_k: float) -> None:
        """
        creates all valid joint degrees for overall k and their probability
        and updates self._jdd.
        :param k: overall degree
        :param prob_overall_k: float value
        """
        self._jdd = {}

        # get a list of valid joint degrees
        valid_tuples: list = list(self.get_valid_joint_degrees(k, len(self._probs)))

        # calculate the probability of each joint degree tuple
        probabilities = []
        for jd in valid_tuples:
            probabilities.append(self.calc_prob_of_joint_degree(jd))

        # normalise the probabilities to unity
        total = sum(probabilities)
        for i in range(len(probabilities)):
            probabilities[i] /= total

        # add each tuple to self._jdd weighted by probability of overall degree k
        for i, jd in enumerate(valid_tuples):
            self._jdd[tuple(jd)] = prob_overall_k * probabilities[i]



from .joint_degree import JointDegree
from .joint_degree_type import JointDegreeType
from .joint_degree_loaders.joint_degree_manual import (
    JointDegreeManual,
)
from .joint_degree_loaders.joint_degree_empirical import (
    JointDegreeEmpirical,
)
from .joint_degree_loaders.joint_degree_function import (
    JointDegreeFunction,
)
from .joint_degree_loaders.joint_degree_marginal import (
    JointDegreeMarginal,
)
from .joint_degree_loaders.joint_degree_split_degree import (
    JointDegreeSplitDegree,
)
from .joint_degree_loaders.joint_degree_delta import JointDegreeDelta
from .joint_degree_loaders.joint_degree_cover import JointDegreeCover


class JointDegreeFactory:
    @staticmethod
    def resolve_joint_degree(type: JointDegreeType, params: dict) -> JointDegree:
        if type == JointDegreeType.MANUAL:
            return JointDegreeManual(params)
        elif type == JointDegreeType.EMPIRICAL:
            return JointDegreeEmpirical(params)
        elif type == JointDegreeType.JOINT_FUNCTION:
            return JointDegreeFunction(params)
        elif type == JointDegreeType.MARGINAL:
            return JointDegreeMarginal(params)
        elif type == JointDegreeType.SPLIT_DEGREE:
            return JointDegreeSplitDegree(params)
        elif type == JointDegreeType.DELTA:
            return JointDegreeDelta(params)
        elif type == JointDegreeType.COVER:
            return JointDegreeCover(params)
        else:
            raise ValueError("Unknown joint degree type")



class JointExcessJointDegreeKeysView:

    """
    Utility class to access correct view of the joint
    excess joint degree keys.
    :param keys: ordered list of keys [u0, u1, v0, v1]
    """

    def __init__(self, keys: list):
        self._keys: list = keys

    def get_u0u1(self):
        return self._keys[0] + self._keys[1]

    def get_u1u0(self):
        return self._keys[1] + self._keys[0]

    def get_v0v1(self):
        return self._keys[2] + self._keys[3]

    def get_v1v0(self):
        return self._keys[3] + self._keys[2]

    def get_u0v1(self):
        return self._keys[0] + self._keys[3]

    def get_v0u1(self):
        return self._keys[2] + self._keys[1]



from gcmpy.joint_degree.joint_degree_loaders.joint_degree_split_degree import (
    JointDegreeSplitDegree,
)
from gcmpy.joint_degree.joint_degree_type import JointDegreeType
from gcmpy.names.joint_degree_names import JointDegreeNames


class JointDegreeDelta(JointDegreeSplitDegree):

    _type: str = JointDegreeType.DELTA

    def __init__(self, params: dict):

        try:
            self._target_k = params[JointDegreeNames.TARGET_K]
            self._fp = params[JointDegreeNames.FP]
            self._probs = params[JointDegreeNames.PROBS]
            self._motif_sizes = params[JointDegreeNames.MOTIF_SIZES]
            self._low_high_degree_bound = params[JointDegreeNames.LOW_HIGH_DEGREE_BOUND]
        except Exception as e:
            raise (f"Error instantiating {self.__class__.__name__}: {e}")

        self.create_jdd()

    def create_jdd(self) -> None:
        self._jdd = {}
        for k in range(self._low_high_degree_bound[0], self._low_high_degree_bound[1]):
            zeros = [0] * len(self._motif_sizes)
            if k != self._target_k:
                zeros[0] = k
                self._jdd[tuple(zeros)] = self._fp(k)
            else:
                self.resolve_degree(k, self._fp(k))
        self.normalise_jdd()
